
'''Various functions that are serve I/O purposes or not relevant to any specific module.

'''

#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from math import cos, sqrt, pi
from haversine import haversine
import numpy as np
import pandas as pd
from os import path
import dill as pickle
from compress_pickle import dump, load
from vincenty import vincenty


ptB = [145.72726054, -16.81220888]
ptA = [145.7274588, -16.81253358]

###############################################################################
# Vincenty distance between points
###############################################################################
def vincentyDistance(pointA, pointB, meters=True):
    """Calculates the Vincenty arc distance between points.
    
    Args:
        pointA (tuple): First point in (lon, lat) format.
        pointB (tuple): Second point in (lon, lat) format.

    Returns:
        (float): Distance between points
    """
    distKM = vincenty(
        (pointA[1], pointA[0]), 
        (pointB[1], pointB[0])
    )
    if meters:
        return distKM*1000
    else:
        return distKM


def cheapRuler(pointA, pointB):
    """Calculates the distance between two (lon,lat) points assumming flat approximations.
        https://blog.mapbox.com/fast-geodesic-approximations-with-cheap-ruler-106f229ad016
    """

    """
    (parLen, merLen) = (40074.275e3, 20004.146e3)
    (lonA, latA) = pointA
    (lonB, latB) = pointB
    # Distance calculations
    (dx, dy) = (
        parLen*(abs(lonA-lonB)/360)*cos((latA+latB)/2),
        merLen*(abs(latA-latB)/180)
    )
    distance = sqrt(dx**2+dy**2)
    """

    """Following code based on:
        https://github.com/mapbox/cheap-ruler/blob/master/index.js
    """
    (lonA, latA) = pointA
    (lonB, latB) = pointB

    RE = 6378.137
    FE = 1 / 298.257223563
    E2 = FE * (2 - FE)
    RAD = pi / 180

    m = RAD * RE * 1000
    coslat = cos(latA * RAD)
    w2 = 1 / (1- E2 * (1 - coslat * coslat))
    w = sqrt(w2)

    kx = m * w * coslat
    ky = m * w * w2 * (1 - E2)

    deg = lonA - lonB

    while (deg < -180): 
        deg += 360
    while (deg > 180): 
        deg -= 360

    dx = deg * kx
    dy = (latA - latB) * ky
    distance = sqrt(dx**2+dy**2)

    return distance


def haversineDistance(pointA, pointB):
    """Calculates the distance between two (lon,lat) points using the haversine approximation.

    """
    (lonA, latA) = pointA
    (lonB, latB) = pointB
    distance = haversine((latB, lonB), (latA, lonA), unit='m')
    return distance

###############################################################################
# Various auxiliary functions
###############################################################################
def makeFolder(path):
    """Creates a folder if it doesn't already exist.
    
    Args:
        path (string): Path to the desired directory.
    """
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except OSError:
            raise OSError(
                    "Can't create destination directory (%s)!" % (path)
                )

def makeFolders(pathsList):
    """Creates a list of folders if they don't exist.
    
    Args:
        paths (list): List path to the desired directories.
    """
    for fldr in pathsList:
        makeFolder(fldr)


def isNotebook():
    """Checks if the script is running from a Jupyter environment.

    Returns:
        bool: Flags Jupyter environment. 
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

###############################################################################
# Dump/Load functions
###############################################################################
def dumpLandscape(landscape, fPath, fName, fExt='bz2'):
    """Exports a serialized landscape to disk.

    Args:
        landscape (object): Landscape object to export.
        fPath (path): Path to where the landscape will be exported.
        fName (string): Filename.
        fExt (string): File extension.
    """
    if fExt == 'bz2':
        dump(
            landscape, 
            path.join(fPath, '{}.{}'.format(fName, fExt))
        )
    else:
        with open(path.join(fPath, '{}.{}'.format(fName, fExt)), 'wb') as f:
            pickle.dump(landscape, f)
        

def loadLandscape(fPath, fName, fExt='bz2'):
    """Loads a serialized landscape from disk.

    Args:
        fPath (path): Path from where the landscape will be loaded.
        fName (string): Filename.
        fExt (string): File extension.

    Returns:
        (object): Landscape object.
    """
    if fExt == 'bz2':
        lnd = load(
            path.join(fPath, '{}.{}'.format(fName, fExt))
        )
    else:
        with open(path.join(fPath, '{}.{}'.format(fName, fExt)), 'rb') as f:
            lnd = pickle.load(f)
    return lnd

def exportLandscape(landscape, fPath, fName):
    """Exports a landscape to disk to CSV files (for use in MGDrivE).

    Args:
        landscape (object): Landscape object to export.
        fPath (path): Path to where the landscape will be exported.
        fName (string): Filename.

    """
    coords = np.concatenate(
        (landscape.pointCoords, landscape.trapsCoords), 
        axis=0
    )
    sitesTypes = np.concatenate(
        (landscape.pointTypes, np.asarray(landscape.trapsNumber*[-1])),
        axis=0
    )
    trapTypes = np.concatenate(
        (np.asarray(landscape.pointNumber*[-1]), landscape.trapsTypes),
        axis=0
    )
    coordsT = coords.T
    outArray = np.vstack((coordsT[0], coordsT[1], sitesTypes, trapTypes)).T
    outDF = pd.DataFrame(
        outArray, columns=('x', 'y', 'SitesType', 'TrapsType')
    )
    # Export to disk
    outDF.to_csv(
        path.join(fPath, '{}_Sites.csv'.format(fName)),
        index=False
    )
    np.savetxt(
        path.join(fPath, '{}_Migration.csv'.format(fName)),
        landscape.trapsMigration, 
        delimiter=","
    )

###############################################################################
# Paths functions
###############################################################################
def makeFolder(path):
    """Crates a folder in the specified directory.

    Args:
        path (string): Path of the folder than needs to be created.

    """
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except OSError:
            raise OSError(
                    "Can't create destination directory (%s)!" % (path)
                )


'''Synthetic Landscape generation functions.

'''

import math
import time
import random
import numpy as np
import numpy.random as rand
from sklearn.cluster import KMeans
import libpysal as ps
from libpysal.cg import shapely_ext
from pointpats import PoissonClusterPointProcess, Window


def ptsRegularGrid(pointsNumber, bbox):
    """ Creates a regular grid (lattice) of points.
    
    Parameters:
        pointsNumber (int): Number of sites.
        bbox (tuple of tuples): Bounding box in the form ((xLo, xHi), (yLo, yHi)).

    Returns:
        (numpy array): Points' coordinates.
    """
    (xRan, yRan) = bbox
    x = np.linspace(xRan[0], xRan[1], pointsNumber)
    y = np.linspace(yRan[0], yRan[1], pointsNumber)
    coords = np.asarray(np.meshgrid(x, y)).T
    coords = np.concatenate(coords)
    return coords


def ptsDonut(pointsNumber, radii, center=(0, 0)):
    """ Creates a distribution of points laid around a donut shape.
    
    Parameters:
        pointsNumber (int): Number of sites.
        radii (tuple of floats): Minimum and maximum radii for the donut shape.
        center (tuple of floats): Coordinates for the center of the donut (x, y).

    Returns:
        (numpy array): Points' coordinates.
    """
    coords = []
    for i in range(pointsNumber):
        radius = (radii[0], radii[1])
        angle = 2 * math.pi * random.random()
        # random radius
        r = rand.uniform(radius[0], radius[1])
        # calculating coordinates
        x = r * math.cos(angle) + center[0]
        y = r * math.sin(angle) + center[1]
        coords.append([x, y])
    return np.asarray(coords)


def ptsRandUniform(pointsNumber, bbox):
    """ Creates a random unifor distribution of points.
    
    Parameters:
        pointsNumber (int): Number of sites
        bbox (tuple of tuples): Bounding box in the form ((xLo, xHi), (yLo, yHi))

    Returns:
        (numpy array): Points' coordinates
    """
    (xRan, yRan) = bbox
    xy = (
        rand.uniform(*xRan, pointsNumber), 
        rand.uniform(*yRan, pointsNumber)
    )
    coords = list(zip(*xy))
    return np.asarray(coords)


def ptsPossion(
    pointsNumber, clustersNumber, 
    radius, randomState=time.time(), 
    bbox=None, polygon=None
    ):
    """Generates a synthetic landscape from a Poisson distribution.
    
    Parameters:
        pointsNumber (int): Number of sites.
        clustersNumber (int): Number of sites' clusters.
        radius (float): Radius of the circle centered on each parent.
        randomState (int): Random seed.

        bbox (tuple of tuples): Bounding box in the form ((xLo, xHi), (yLo, yHi))
        polygon (name of shp file): Shape-file for the points to be generated within.

    Returns:
        (numpy array):  Points' coordinates
    """

    if bbox:
        ((xLo, xHi), (yLo, yHi)) = (bbox[0], bbox[1])
        parts = [[(xLo, yLo), (xLo, yHi), (xHi, yHi), (xHi, yLo), (xLo, yLo)]]
        window = Window(parts)
    elif polygon:
        shp_file = ps.io.open(polygon)
        polys = [shp for shp in shp_file]
        state = shapely_ext.cascaded_union(polys)
        window = Window(state.parts)

    np.random.seed(int(randomState))
    csamples = PoissonClusterPointProcess(
        window, pointsNumber, 
        clustersNumber, radius, 1, asPP=True, conditioning=False
    )

    return np.asarray(csamples.realize(pointsNumber))


###############################################################################
# Clustering and Aggregating Landscape
###############################################################################

# def clusterLandscape(
#         pointsCoords, clustersNumber, 
#         randomState=time.time(), clusterAlgorithm=KMeans
#     ):
#     """ .
    
#     Parameters:
#         pointsCoords (np array): 
#         clustersNumber (int):
#         randomState (int):
#         clusterAlgorithm (function):

#     Returns:
#         (dict): 
#     """
#     clObj = clusterAlgorithm(
#         n_clusters=clustersNumber,
#         random_state=int(randomState)
#     )
#     clustersObj = clObj.fit(pointsCoords)
#     (clusters, centroids) = (
#         clustersObj.labels_,
#         clustersObj.cluster_centers_
#     )
#     return {'clusters': clusters, 'centroids': centroids}


# def aggregateLandscape(migrationMatrix, clusters):
#     """ .
    
#     Parameters:
#         migrationMatrix (np matrix): 
#         clusters (list): 

#     Returns:
#         (numpy array): 
#     """
#     matrix_size = len(clusters)
#     num_clusters = len(set(clusters))
#     aggr_matrix = np.zeros([num_clusters, num_clusters], dtype=float)
#     aggr_number = [0]*num_clusters
#     for row in range(matrix_size):
#         cRow = clusters[row]
#         aggr_number[cRow] += 1
#         for col in range(matrix_size):
#             cCol = clusters[col]
#             aggr_matrix[cRow][cCol] += migrationMatrix[row][col]
#     for row in range(num_clusters):
#         aggr_matrix[row] = [x/aggr_number[row] for x in aggr_matrix[row]]
#     return aggr_matrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sys import argv
import math
import numpy as np
import pandas as pd
from os import path
from sys import argv
from copy import deepcopy
import matplotlib.pyplot as plt
from deap import base, creator, algorithms, tools
from compress_pickle import dump, load
from sklearn.preprocessing import normalize
import MGSurvE as srv
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import warnings
warnings.filterwarnings('ignore', 'The iteration is not making good progress')


MAIL_ALERTS = True
(FXD_TRPS, TRPS_NUM) = (int(argv[2]), int(argv[1]))
# (FXD_TRPS, TRPS_NUM) = (True, 8)
###############################################################################
# Debugging fixed traps at land masses
###############################################################################
OUT_PTH = '/RAID5/marshallShare/MGS_Benchmarks/STPVincenty/'
# OUT_PTH = '/home/chipdelmal/Documents/WorkSims/MGSurvE_Benchmarks/STPVincenty'
if FXD_TRPS:
    ID = 'STP_FXD'
else:
    ID = 'STP_FXN'
GENS = 5000
(IX_SPLIT, DIAG_VAL) = (27, 0.02)
###############################################################################
# Setup email alerts
###############################################################################
if MAIL_ALERTS:
    import time
    import smtplib
    import mailAlerts as mlr
    from email.message import EmailMessage
    import imghdr
    t0 = time.time()
###############################################################################
# Load Pointset
###############################################################################
sites = pd.read_csv(path.join(OUT_PTH, 'GEO', 'stp_cluster_sites_pop_v5_fixed.csv'))
sites['t'] = [0]*sites.shape[0]
SAO_TOME_LL = sites.iloc[IX_SPLIT:]
SAO_bbox = (
    (min(SAO_TOME_LL['lon']), max(SAO_TOME_LL['lon'])),
    (min(SAO_TOME_LL['lat']), max(SAO_TOME_LL['lat']))
)
SAO_cntr = [i[0]+(i[1]-i[0])/2 for i in SAO_bbox]
SAO_TOME_LL = SAO_TOME_LL .rename(
    columns={'lon': 'x', 'lat': 'y'}
)
SAO_LIMITS = ((6.41, 6.79), (-0.0475, .45))
# Get location of minor land-masses -------------------------------------------
SAO_FIXED = [tuple(SAO_TOME_LL.loc[i][['x', 'y']]) for i in (51, 239)]
FXD_NUM = len(SAO_FIXED)
###############################################################################
# Load Migration Matrix
###############################################################################
migration = np.genfromtxt(
    path.join(OUT_PTH, 'GEO', 'kernel_cluster_v6a.csv'), delimiter=','
)
msplit = migration[IX_SPLIT:,IX_SPLIT:]
np.fill_diagonal(msplit, DIAG_VAL)
SAO_TOME_MIG = normalize(msplit, axis=1, norm='l1')
###############################################################################
# Defining Traps
#   North and South land masses (mini islands): 51, 239 (zero-indexed)
###############################################################################
(initTyp, initFxd) = ([0]*TRPS_NUM, [0]*TRPS_NUM)
(initLon, initLat) = ([SAO_cntr[0]]*TRPS_NUM, [SAO_cntr[1]]*TRPS_NUM)
if FXD_TRPS:
    for i in range(FXD_NUM):
        initFxd[TRPS_NUM-(i+1)] = 1
        initLon[TRPS_NUM-(i+1)] = SAO_FIXED[i][0]
        initLat[TRPS_NUM-(i+1)] = SAO_FIXED[i][1]
traps = pd.DataFrame({'x': initLon, 'y': initLat, 't': initTyp, 'f': initFxd})
tKer = {0: {'kernel': srv.exponentialDecay, 'params': {'A': .5, 'b': 100}}}
###############################################################################
# Setting Landscape Up
###############################################################################
lnd = srv.Landscape(
    SAO_TOME_LL, migrationMatrix=SAO_TOME_MIG,
    traps=traps, trapsKernels=tKer, landLimits=SAO_LIMITS
)
bbox = lnd.getBoundingBox()
trpMsk = srv.genFixedTrapsMask(lnd.trapsFixed)
###############################################################################
# Plot Landscape
###############################################################################
# (fig, ax) = (
#     plt.figure(figsize=(15, 15)),
#     plt.axes(projection=ccrs.PlateCarree())
# )
# lnd.plotSites(fig, ax, size=250)
# lnd.plotTraps(fig, ax)
# lnd.plotMigrationNetwork(
#     fig, ax, 
#     lineWidth=10, alphaMin=.1, alphaAmplitude=2.5,
# )
# lnd.plotLandBoundary(fig, ax)
# srv.plotClean(fig, ax, bbox=lnd.landLimits)
# fig.savefig(
#     path.join(OUT_PTH, '{}_{:02d}_CLN.png'.format(ID, TRPS_NUM)), 
#     facecolor='w', bbox_inches='tight', pad_inches=0.1, dpi=300
# )
# plt.close('all')
###############################################################################
# GA Settings
############################################################################### 
POP_SIZE = int(10*(lnd.trapsNumber*1.25))
(MAT, MUT, SEL) = (
    {'mate': .35, 'cxpb': 0.5}, 
    {
        'mean': 0, 
        'sd': max([abs(i[1]-i[0]) for i in bbox])/5, 
        'mutpb': .35, 'indpb': .5
    },
    {'tSize': 5}
)
VERBOSE = True
lndGA = deepcopy(lnd)
# Reducing the bbox for init sampling -----------------------------------------
redFract = 0
reduction = [(i[1]-i[0])/2*redFract for i in bbox]
bboxRed = [(i[0]+r, i[1]-r) for (i, r) in zip(bbox,reduction)]
###############################################################################
# Registering GA functions
############################################################################### 
toolbox = base.Toolbox()
creator.create(
    "FitnessMin", base.Fitness, 
    weights=(-1.0, )
)
creator.create(
    "Individual", list, 
    fitness=creator.FitnessMin
)
toolbox.register(
    "initChromosome", srv.initChromosome, 
    trapsCoords=lndGA.trapsCoords, 
    fixedTrapsMask=trpMsk, coordsRange=bboxRed
)
toolbox.register(
    "individualCreator", tools.initIterate, 
    creator.Individual, toolbox.initChromosome
)
toolbox.register(
    "populationCreator", tools.initRepeat, 
    list, toolbox.individualCreator
)
# Mate and mutate -------------------------------------------------------------
if FXD_TRPS:
    toolbox.register(
        "mutate", srv.mutateChromosome, 
        fixedTrapsMask=trpMsk, indpb=MUT['indpb'],
        randArgs={'loc': MUT['mean'], 'scale': MUT['sd']}
    )
    toolbox.register(
        "mate", srv.cxBlend, 
        fixedTrapsMask=trpMsk, alpha=MAT['mate']
    )
else:
    toolbox.register(
        "mutate", tools.mutGaussian, 
        mu=MUT['mean'], sigma=MUT['sd'], indpb=MUT['indpb']
    )
    toolbox.register(
        "mate", tools.cxBlend, 
        alpha=MAT['mate']
    )
# Select and evaluate ---------------------------------------------------------
toolbox.register(
    "select", tools.selTournament, 
    tournsize=SEL['tSize']
)
toolbox.register(
    "evaluate", srv.calcFitness, 
    landscape=lndGA,
    optimFunction=srv.getDaysTillTrapped,
    optimFunctionArgs={'outer': np.mean, 'inner': np.max}
)
###############################################################################
# Registering GA stats
############################################################################### 
pop = toolbox.populationCreator(n=POP_SIZE)
hof = tools.HallOfFame(1)
stats = tools.Statistics(lambda ind: ind.fitness.values)   
stats.register("min", np.min)
stats.register("avg", np.mean)
stats.register("max", np.max)
stats.register("best", lambda fitnessValues: fitnessValues.index(min(fitnessValues)))
stats.register("traps", lambda fitnessValues: pop[fitnessValues.index(min(fitnessValues))])
###############################################################################
# Optimization Cycle
############################################################################### 
(pop, logbook) = algorithms.eaSimple(
    pop, toolbox, 
    cxpb=MAT['cxpb'], mutpb=MUT['mutpb'], 
    ngen=GENS, stats=stats, halloffame=hof, verbose=VERBOSE
)
###############################################################################
# Get and Export Results
############################################################################### 
bestChromosome = hof[0]
bestTraps = np.reshape(bestChromosome, (-1, 2))
lnd.updateTrapsCoords(bestTraps)
srv.dumpLandscape(lnd, OUT_PTH, '{}_{:02d}_TRP'.format(ID, TRPS_NUM))
dta = pd.DataFrame(logbook)
srv.exportLog(logbook, OUT_PTH, '{}_{:02d}_LOG'.format(ID, TRPS_NUM))
###############################################################################
# Plot Results
###############################################################################
(fig, ax) = (
    plt.figure(figsize=(15, 15)),
    plt.axes(projection=ccrs.PlateCarree())
)
lnd.plotSites(fig, ax, size=250)
lnd.plotMigrationNetwork(
    fig, ax, lineWidth=10, alphaMin=.1, alphaAmplitude=2.5
)
lnd.plotTraps(fig, ax, zorders=(25, 20))
srv.plotFitness(fig, ax, min(dta['min']), fmt='{:.2f}')
lnd.plotLandBoundary(fig, ax)
srv.plotClean(fig, ax, bbox=lnd.landLimits)
fig.savefig(
    path.join(OUT_PTH, '{}_{:02d}_TRP.png'.format(ID, TRPS_NUM)), 
    facecolor='w', bbox_inches='tight', pad_inches=0.1, dpi=400
)
plt.close('all')
###############################################################################
# Email Finished!
############################################################################### 
if MAIL_ALERTS:
    tF = (time.time()-t0)/3600
    fName = __file__.split('/')[-1]
    mailStr = 'Sim finished ({}  t{:02d})\nRuntime: {}'.format(
        fName, TRPS_NUM, tF
    )
    msg = EmailMessage()
    msg['Subject'] = 'Sim finished: {} - {:02d}'.format(fName, TRPS_NUM)
    msg['From'] = mlr.MAIL
    msg['To'] = mlr.TARG
    msg.set_content(mailStr)
    imgPath = path.join(OUT_PTH, '{}_{:02d}_TRP.png'.format(ID, TRPS_NUM))
    with open(imgPath, 'rb') as f:
        image_data = f.read()
        image_type = imghdr.what(f.name)
        image_name = f.name
    msg.add_attachment(
        image_data, maintype='image', subtype='image_type', filename=image_name
    )
    with smtplib.SMTP_SSL('smtp.gmail.com', 465) as smtp:
        smtp.login(mlr.MAIL, mlr.PSWD)
        smtp.send_message(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import setuptools
from version import version as this_version


this_directory =  os.path.abspath(os.path.dirname(__file__))
version_path = os.path.join(this_directory, 'MGSurvE', '_version.py')
print(this_version)
with open(version_path, 'wt') as fversion:
    fversion.write('__version__ = "'+this_version+'"')

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='MGSurvE',
    version=this_version,
    url='https://github.com/Chipdelmal/MGSurvE',
    author='Hector M. Sanchez C.',
    author_email='sanchez.hmsc@berkeley.edu',
    description="MGSurvE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'deap', 'numpy', 'scikit-learn', 'scipy', 'matplotlib', 
        'ipython', 'jupyter', 'pandas', 'compress-pickle', 'dill', 
        'vincenty', 'haversine', 'networkx', 'pointpats', 'libpysal',
        'geopandas'
    ],
    extras_require={
        'dev': ['pytest', 'twine', 'wheel', 'sphinx', 'sphinx_rtd_theme'],
    },
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3.9',
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ]
 )